import asyncio
import concurrent
from functools import partial

import peewee
from playhouse.db_url import connect

from .serialize import Serializer
from .migrate import Router
from muffin.plugins import BasePlugin, PluginException


class PeeweePlugin(BasePlugin):

    """ Integrate peewee to bottle. """

    name = 'peewee'
    defaults = {
        'connection': 'sqlite:///db.sqlite',
        'max_connections': 2,
        'migrations_path': 'migrations',
    }

    def __init__(self, **options):
        super().__init__(**options)

        self.database = peewee.Proxy()
        self.serializer = Serializer()
        self.models = dict()

    def setup(self, app):
        """ Initialize the application. """
        super().setup(app)
        if 'manage' not in app.plugins:
            raise PluginException('Peewee plugin requires Manage plugin initialized before.')

        # Setup Database
        self.database.initialize(connect(self.options['connection']))
        self.threadpool = concurrent.futures.ThreadPoolExecutor(
            max_workers=self.options['max_connections'])

        # Setup migration engine
        self.router = Router(self)

        # Register migration commands
        @self.app.ps.manage.command
        def migrate(name:str=None):
            """ Run application's migrations.

            :param name: Choose a migration' name

            """
            self.router.run(name)

        @self.app.ps.manage.command
        def create(name:str):
            """ Create a migration.

            :param name: Set name of migration [auto]

            """
            self.router.create(name)

    @asyncio.coroutine
    def middleware_factory(self, app, handler):
        """ Control connection to database. """
        @asyncio.coroutine
        def middleware(request):
            if self.options['connection'].startswith('sqlite'):
                return (yield from handler(request))

            self.database.connect()
            response = yield from handler(request)
            if not self.database.is_closed():
                self.database.close()
            return response

        return middleware

    def query(self, query):
        if isinstance(query, peewee.SelectQuery):
            return self.run(lambda: list(query))
        return self.run(query.execute)

    @asyncio.coroutine
    def run(self, function, *args, **kwargs):
        """ Run sync code asyncronously. """
        if kwargs:
            function = partial(function, **kwargs)

        def iteration(database, *args):
            database.connect()
            try:
                with database.transaction():
                    return function(*args)
            except peewee.PeeweeException:
                database.rollback()
                raise
            finally:
                database.commit()

        return (
            yield from self.app.loop.run_in_executor(
                self.threadpool, iteration, self.database,  *args))

    def to_dict(self, obj, **kwargs):
        return self.serializer.serialize_object(obj, **kwargs)

    def register(self, model):
        """ Register a model in self. """
        self.models[model._meta.db_table] = model
        model._meta.database = self.database
        return model



import asyncio
import concurrent
import threading

import peewee


class AsyncDatabaseMixin:

    max_connections = 4

    def __init__(self, database, loop=None, **kwargs):
        self.connections = {}
        super().__init__(database, **kwargs)
        self.loop = loop or asyncio.get_event_loop()
        self.threadpool = concurrent.futures.ThreadPoolExecutor(max_workers=self.max_connections)

    @property
    def _Database__local(self):
        tid = threading.get_ident()
        return self.connections.get(tid, peewee._ConnectionLocal())

    @_Database__local.setter
    def _Database__local(self, value):
        tid = threading.get_ident()
        self.connections[tid] = value



""" Implement Muffin Application. """
import asyncio
import logging
import re
import os
from importlib import import_module
from types import FunctionType

from aiohttp import web
from cached_property import cached_property

from .handler import Handler
from .utils import Structure


CONFIGURATION_ENVIRON_VARIABLE = 'MUFFIN_CONFIG'
RETYPE = type(re.compile('@'))


class MuffinException(Exception):

    """ Implement a Muffin's exception. """

    pass


class RawRoute(web.DynamicRoute):

    def url(self, *, parts, query=None):
        return None


# FIXME: See https://github.com/KeepSafe/aiohttp/pull/291
class Router(web.UrlDispatcher):

    """ Support raw regexps in path. """

    def add_route(self, method, path, handler, *, name=None):
        assert callable(handler), handler
        if not asyncio.iscoroutinefunction(handler):
            handler = asyncio.coroutine(handler)
        method = method.upper()
        assert method in self.METHODS, method

        if isinstance(path, RETYPE):
            route = RawRoute(method, handler, name, path, path.pattern)
            self._register_endpoint(route)
            return route

        return super(Router, self).add_route(method, path, handler, name=name)


class Application(web.Application):

    """ Improve aiohttp Application. """

    # Default application settings
    __defaults = {

        # Configuration module
        'CONFIG': 'config',

        # Enable debug mode
        'DEBUG': False,

        # Install the plugins
        'PLUGINS': (
            'muffin.plugins.manage:ManagePlugin',
            'muffin.plugins.jade:JadePlugin',
            'muffin.plugins.peewee:PeeweePlugin',
            'muffin.plugins.session:SessionPlugin',
        ),

        # Setup static files in development
        'STATIC_PREFIX': '/static',
        'STATIC_ROOT': 'static',
    }

    def __init__(self, name, *, loop=None, router=None, middlewares=(), logger=web.web_logger,
                 handler_factory=web.RequestHandlerFactory, **OPTIONS):
        """ Initialize the application. """
        router = router or Router()

        super(Application, self).__init__(loop=loop, router=router, middlewares=middlewares,
                                          logger=logger, handler_factory=handler_factory)

        self.name = name
        self.defaults = dict(self.__defaults)
        self.defaults.update(OPTIONS)

        self._middlewares = list(self._middlewares)
        self._start_callbacks = []

        # Setup logging
        ch = logging.StreamHandler()
        ch.setFormatter(logging.Formatter(
            '%(asctime)s [%(process)d] [%(levelname)s] %(message)s',
            '[%Y-%m-%d %H:%M:%S %z]'))
        self.logger.addHandler(ch)
        self.logger.setLevel('DEBUG') if self.cfg.DEBUG else self.logger.setLevel('WARNING')
        self.logger.name = 'muffin'

        # Setup plugins
        self.plugins = self.ps = Structure()
        for plugin in self.cfg.PLUGINS:
            try:
                self.install(plugin)
            except Exception as exc:
                self.logger.error('Plugin is invalid: %s (%s)' % (plugin, exc))

        # Setup static files (development)
        if os.path.isdir(self.cfg.STATIC_ROOT):
            self.router.add_static(self.cfg.STATIC_PREFIX, self.cfg.STATIC_ROOT)
        else:
            self.logger.warn('Disable STATIC_ROOT (hasnt found): %s' % self.cfg.STATIC_ROOT)

    def __call__(self, *args, **kwargs):
        return self

    @cached_property
    def cfg(self):
        """ Load the application configuration. """
        config = Structure(self.defaults)
        module = config['CONFIG'] = os.environ.get(
            CONFIGURATION_ENVIRON_VARIABLE, config['CONFIG'])
        try:
            module = import_module(module)
            config.update({
                name: getattr(module, name) for name in dir(module)
                if name == name.upper() and not name.startswith('_')
            })
        except ImportError:
            self.logger.warn("The configuration hasn't found: %s" % module)

        return config

    def install(self, plugin):
        """ Install plugin to the application. """

        if isinstance(plugin, str):
            module, attr = plugin.split(':')
            module = import_module(module)
            plugin = getattr(module, attr)

        if isinstance(plugin, type):
            plugin = plugin()

        if hasattr(plugin, 'setup'):
            plugin.setup(self)

        if hasattr(plugin, 'middleware_factory') \
                and plugin.middleware_factory not in self.middlewares:
            self.middlewares.append(plugin.middleware_factory)

        if hasattr(plugin, 'start'):
            self.register_on_start(plugin.start)

        self.plugins[plugin.name] = plugin

    @asyncio.coroutine
    def start(self):
        """ Start application. """
        for (cb, args, kwargs) in self._start_callbacks:
            try:
                res = cb(self, *args, **kwargs)
                if (asyncio.iscoroutine(res) or isinstance(res, asyncio.Future)):
                    yield from res
            except Exception as exc:
                self._loop.call_exception_handler({
                    'message': "Error in start callback",
                    'exception': exc,
                    'application': self,
                })

    def register_on_start(self, func, *args, **kwargs):
        self._start_callbacks.append((func, args, kwargs))

    def register(self, *paths, methods=['GET'], name=None):
        """ Register function (coroutine) or muffin.Handler to application and bind to route. """

        if isinstance(methods, str):
            methods = [methods]

        def wrapper(view):
            handler = view

            if isinstance(handler, FunctionType):
                handler = Handler.from_view(handler, *methods, name=name)

            handler.connect(self, *paths, name=name)

            return view

        return wrapper


def run():
    """ Run the Gunicorn Application. """
    from .worker import GunicornApp

    GunicornApp("%(prog)s [OPTIONS] [APP]").run()


if __name__ == '__main__':
    run()
